#include <DigitalOutputCard.hpp>
#include <DigitalInputCard.hpp>

/**
 * @brief Configuration structure for the Gasoline Generator.
 *
 * This structure holds the configuration parameters for the gasoline generator,
 * including input and output cards, pin assignments, and sensor settings.
 * It is used to initialize the GasolineGenerator class.
 * * @note The `inputCard` and `outputCard` should be initialized before passing to the GasolineGenerator.
 * * @note The `ignitionSystemPowerRelayPin`, `starterRelayPin`, and `carburetorValvePin` should be set to the correct GPIO pin numbers.
 * * @note The `carburetorValveReverse` flag indicates whether the carburetor valve operates in reverse logic (0=open, 1=closed).
 **/

struct GasolineGeneratorConfig
{
    uint8_t ignitionSystemPowerRelayPin; // Pin for ignition system power relay
    uint8_t starterRelayPin;          // Pin for starter relay
    uint8_t carburetorValvePin;      // Pin for carburetor control
    bool carburetorValveReverse;    // If true, 0 means open, 1 means closed else 0 means closed, 1 means open
    uint8_t powerOutputSensorPin; // Pin for monitoring generator power output
    bool usePowerOutputSensor = false; // Enable/disable power output sensor monitoring
    uint8_t togglePin; // Pin for toggle switch to turn generator on/off
    bool useTogglePin = false; // Enable/disable toggle pin functionality
};

/**
 * @brief Enumeration for the generator states.
 *  This enum defines the possible states of the gasoline generator.
 */
enum class GeneratorState
{
    STOPPED,
    STARTING,
    RUNNING,
    SHUTTING_DOWN
};

/**
 * @brief Gasoline Generator class for managing a gasoline-powered generator.
 *
 * This class handles the ignition system, starter relay, and carburetor control
 * For startup first it closes the carburetor valve, creating a rich air-fuel mixture,
 * then it powers the ignition system, wait a bit then spin the starter relay to start the engine.
 * After the engine starts, it opens the carburetor valve to allow normal operation.
 */
class GasolineGenerator
{
public:
    GasolineGenerator();
    void initialize(DigitalInputCard *inputCard, DigitalOutputCard *outputCard, const GasolineGeneratorConfig &config);
    void loop();
    void start();
    void shutdown();

private:
    const GasolineGeneratorConfig *config = nullptr;
    DigitalInputCard *inputCard = nullptr;
    DigitalOutputCard *outputCard = nullptr;
    GeneratorState state = GeneratorState::STOPPED;
    unsigned long startupStartTime = 0;
    unsigned long ignitionStartTime = 0;
    bool isEngineRunning = false;
    bool starterEngaged = false;

    static const unsigned long IGNITION_DELAY_MS = 2000;        // Time to wait after powering ignition system before engaging starter
    static const unsigned long STARTER_DURATION_MS = 5000;      // Time that the starter should be engaged
    static const unsigned long CARBURETOR_OPEN_DELAY_MS = 1000; // Time to wait after engine start for the engine to stabilize before opening the carburetor valve
    static const unsigned long POWER_SENSOR_DEBOUNCE_MS = 500;  // Debounce time for power output sensor

    void setCarburetorValve(bool open);
    void setIgnitionSystem(bool enable);
    void setStarter(bool enable);
    void handleStartupSequence(unsigned long currentTime);
    void onEngineStarted();
    void handleInputChanges(uint8_t pin, bool state);
};
